// Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
// SPDX-License-Identifier: Apache-2.0 OR ISC

// This file is intentionally empty, were AWS-LC installed with symbol prefixes, this file would contain
// macro definitions for the prefixed symbols.
